//-----------------------------------------------------------------------
// Copyright  Microsoft Corporation.  All Rights Reserved.
// This code released under the terms of the 
// Microsoft Public License (MS-PL, http://opensource.org/licenses/ms-pl.html)
//
//  $File: SerialPort.cs $ $Revision: 2 $
//-----------------------------------------------------------------------
//
// Serial Port Service
//
// NOTES:
// The SerialPort object does buffering so there is no need to have a
// read request in progress prior to issuing a write. Simple write/read
// cycles should be sufficient to work with most devices.
//
// This service does not support modem-control pins.
// It also assumes that the connection does not require flow control.
//
// Data errors are logged, but no action is taken.
//
// There is no support for changing the Newline character. The ReadString request
// waits until a Newline is received but does not include it in the string.
// On the other hand, WriteString appends a Newline character to the string.
//
// ReadByteArray can only read a fixed number of bytes. There is no terminating
// character. WriteByteArray writes a specified number of bytes, but this does
// not have to be the same as the length of the array.
//
// The Inter-Character Delay is a feature of this service, not the .NET SerialPort.
// Some devices with slow microcontrollers cannot keep up if you send data at full
// speed and it gets worse the higher the baud rate. To work around this problem,
// you can insert a delay between characters.
//

using System;
using System.Collections.Generic;
using System.ComponentModel;
using Microsoft.Ccr.Core;
using Microsoft.Dss.Core.Attributes;
using Microsoft.Dss.ServiceModel.Dssp;
using Microsoft.Dss.ServiceModel.DsspServiceBase;
using W3C.Soap;
using submgr = Microsoft.Dss.Services.SubscriptionManager;

// Added references
using System.IO.Ports;

namespace ProMRDS.Robotics.Comms
{
	[Contract(Contract.Identifier)]
	[DisplayName("Serial Port Service")]
	[Description("SerialPort service provides access to a Serial (COM) Port")]
    [DssServiceDescription("http://www.promrds.com/SerialPort.htm")]
    // Make sure that we do not deadlock by providing sufficient threads,
    // although this is overkill
    [ActivationSettings(ShareDispatcher = false, ExecutionUnitsPerDispatcher = 4)]
    class SerialPortService : DsspServiceBase
	{
		/// <summary>
		/// Service state
		/// </summary>
		[ServiceState]
        // Add an initial state partner to read the config file
        [InitialStatePartner(Optional = true, ServiceUri = "SerialPort.Config.xml")]
        SerialPortState _state = null;
		
		/// <summary>
		/// Main service operations port
		/// </summary>
		[ServicePort("/SerialPort", AllowMultipleInstances = true)]
		SerialPortOperations _mainPort = new SerialPortOperations();

        /// <summary>
        /// Subscription Manager handles subscriptions/notifications
        /// </summary>
        [SubscriptionManagerPartner]
        submgr.SubscriptionManagerPort _submgrPort = new submgr.SubscriptionManagerPort();

        /// <summary>
        /// The Serial Port object used to send/receive data
        /// </summary>
        SerialPort sp;

        #region Initialization

        /// <summary>
		/// Service constructor
		/// </summary>
		public SerialPortService(DsspServiceCreationPort creationPort)
			: base(creationPort)
		{
		}
		
		/// <summary>
		/// Service start
		/// </summary>
		protected override void Start()
		{
            // Check the state and initialize it if there was no config file
            InitializeState();

            // Add this service to the directory and accept messages
			base.Start();

            // The port can be opened immediately if requested in the config
            if (_state.OpenOnStart)
                OpenPort();
		}

        /// <summary>
        /// Initialize the State if not specified in the config file
        /// </summary>
        private void InitializeState()
        {
            // If no config file, _state will be null
            if (_state == null)
            {
                // Create a new state with default values
                _state = new SerialPortState();
                _state.IsOpen = false;
                _state.Config = new SerialPortConfig();
                _state.Config.PortNumber = 1;
                _state.Config.BaudRate = 57600;
                _state.Config.Parity = SerialPortParity.None;
                _state.Config.DataBits = 8;
                _state.Config.StopBits = SerialPortStopBits.One;
                _state.Config.ReadTimeout = 5000;       // 5 seconds
                _state.Config.WriteTimeout = 5000;      // 5 seconds
                _state.Config.InterCharacterDelay = 0;  // No delay
                _state.LastByteReceived = 0;

                // Save the new State to the config file
                SaveState(_state);
            }
        }

        #endregion

        #region Serial Port Helpers

        /// <summary>
        /// Open the Serial Port
        /// </summary>
        /// <returns></returns>
        protected bool OpenPort()
        {
            string ComPort = "COM" + _state.Config.PortNumber.ToString();

            // Close the port first if already open
            if (sp != null && sp.IsOpen)
            {
                sp.Close();
                sp = null;
                _state.IsOpen = false;
            }

            // Create a new SerialPort object
            sp = new SerialPort(ComPort);

            // Set the parameters for the port
            sp.BaudRate = _state.Config.BaudRate;

            // This simply translates from the "local" enum to the one that
            // .NET uses for the SerialPort
            switch (_state.Config.Parity)
            {
                case SerialPortParity.Even:
                    sp.Parity = Parity.Even;
                    break;
                case SerialPortParity.Mark:
                    sp.Parity = Parity.Mark;
                    break;
                case SerialPortParity.None:
                    sp.Parity = Parity.None;
                    break;
                case SerialPortParity.Odd:
                    sp.Parity = Parity.Odd;
                    break;
                case SerialPortParity.Space:
                    sp.Parity = Parity.Space;
                    break;
                default:
                    sp.Parity = Parity.None;
                    break;
            }

            // Check the DataBits. This code only accepts 7 or 8.
            // Some UARTs can actually handle other lengths, but this is
            // highly unlikely to be required.
            if (_state.Config.DataBits != 7 && _state.Config.DataBits != 8)
                return false;
            sp.DataBits = _state.Config.DataBits;

            // Translate the StopBits too
            switch (_state.Config.StopBits)
            {
                case SerialPortStopBits.None:
                    sp.StopBits = StopBits.One;
                    break;
                case SerialPortStopBits.One:
                    sp.StopBits = StopBits.One;
                    break;
                case SerialPortStopBits.OnePointFive:
                    sp.StopBits = StopBits.OnePointFive;
                    break;
                case SerialPortStopBits.Two:
                    sp.StopBits = StopBits.Two;
                    break;
                default:
                    sp.StopBits = StopBits.One;
                    break;
            }

            // Don't throw away NUL characters because the data might be binary
            sp.DiscardNull = false;

            // Specify 8-bit encoding for data
            sp.Encoding = System.Text.Encoding.UTF8;

            // Set up for Asynch mode if requested
            if (_state.Asynchronous)
            {
                // Set up an Event Handler for received characters that should fire for
                // every character that arrives.
                // NOTE: This handler might NOT be called for every character. It must
                // check the buffer.
                // Also, the handler will NOT be called if there is already a read in
                // progress, e.g. ReadByte, ReadByteArray or ReadString.
                sp.ReceivedBytesThreshold = 1;
                sp.DataReceived += new SerialDataReceivedEventHandler(DataReceivedHandler);
            }
            // Clear the last byte - Only valid in Asynch mode
            _state.LastByteReceived = 0;

            // Set up a handler to log errors on the serial port
            sp.ErrorReceived += new SerialErrorReceivedEventHandler(DataErrorHandler);

            // Specify the timeouts
            // NOTE: It is a good idea to set a timeout so that the service
            // will not hang indefinitely if the comm link fails.
            if (_state.Config.ReadTimeout > 0)
                sp.ReadTimeout = _state.Config.ReadTimeout;
            else
                sp.ReadTimeout = SerialPort.InfiniteTimeout;
            if (_state.Config.WriteTimeout > 0)
                sp.WriteTimeout = _state.Config.WriteTimeout;
            else
                sp.WriteTimeout = SerialPort.InfiniteTimeout;

            // Now try to open the port
            // Notice that there is no try/catch here, but there could be an
            // exception so it has to be handled by the caller
            sp.Open();

            _state.IsOpen = true;

            return true;
        }

        /// <summary>
        /// Close the Serial Port
        /// </summary>
        /// <returns></returns>
        protected bool ClosePort()
        {
            // Only close if currently open
            if (sp != null && sp.IsOpen)
            {
                sp.Close();
                sp = null;
                _state.IsOpen = false;
            }
            return true;
        }

        /// <summary>
        /// Asynchronous Event Handler for serial data
        /// </summary>
        /// <param name="sender">The serial port object</param>
        /// <param name="e">Event Arguments</param>
        void DataReceivedHandler(object sender, SerialDataReceivedEventArgs e)
        {
            int data;
            // Note that this handler might NOT be called for every byte received
            // if the data arrives too quickly. So you have to check the number of
            // bytes available to be read and read them all.
            while (sp.BytesToRead > 0)
            {
                // Read a byte - this will return immediately because we know
                // that there is data available
                data = sp.ReadByte();
                // Post the byte to our own main port so that notifications
                // will be sent out
                ReceiveByte rb = new ReceiveByte();
                rb.Body.Data = (byte)data;
                _mainPort.Post(rb);
                // Set up a Receiver for the response
                Activate(Arbiter.Choice(rb.ResponsePort,
                    success => { },
                    fault => { LogError("ReceiveByte failed"); }
                    ));
            }
        }

        /// <summary>
        /// Asynchronous Event Handler for errors on the serial port
        /// </summary>
        /// <param name="sender">The serial port object</param>
        /// <param name="e">Event arguments - contains the Event type</param>
        void DataErrorHandler(object sender, SerialErrorReceivedEventArgs e)
        {
            LogError("Serial Error: " + e.EventType.ToString());
        }

        #endregion

        #region Operation Handlers

        /// <summary>
        /// Set the serial port configuration parameters
        /// </summary>
        /// <param name="config">Configuration parameters</param>
        /// <remarks>Setting the config while the port is open has no effect.
        /// To change dynamically: close the port, set the config, then open it.
        /// You can use a config file instead of using this operation.
        /// </remarks>
        [ServiceHandler]
        public void SetConfigHandler(SetConfig config)
        {
            if (_state == null)
            {
                _state = new SerialPortState();
                _state.Config = new SerialPortConfig();
            }

            // Do some checking. This should be elaborated.
            SerialPortConfig Config = config.Body.Config;
            if (Config.DataBits != 7 && Config.DataBits != 8)
                throw(new Exception("Invalid Data Bits"));

            // Set all the port config parameters
            _state.Config.PortNumber = Config.PortNumber;
            _state.Config.BaudRate = Config.BaudRate;
            _state.Config.Parity = Config.Parity;
            _state.Config.DataBits = Config.DataBits;
            _state.Config.StopBits = Config.StopBits;
            _state.Config.ReadTimeout = Config.ReadTimeout;
            _state.Config.WriteTimeout = Config.WriteTimeout;
            _state.Config.InterCharacterDelay = Config.InterCharacterDelay;

            // Notify subscribers
            SendNotification(_submgrPort, config);
            // Acknowledge the request
            config.ResponsePort.Post(DefaultUpdateResponseType.Instance);
        }

        /// <summary>
        /// Opens (or re-opens) the serial port
        /// </summary>
        /// <param name="open">The Open request - specifies if asynchronous or not</param>
        [ServiceHandler]
        public void OpenHandler(Open open)
        {
            // Remember the Asynchronous flag
            _state.Asynchronous = open.Body.Asynchronous;
            // Open the serial port
            // If there is an exception, the code will not return and a
            // Fault will be sent back
            if (OpenPort())
            {
                // Notify subscribers
                SendNotification(_submgrPort, open);
                // Respond to the request
                open.ResponsePort.Post(DefaultUpdateResponseType.Instance);
            }
            else
            {
                // Handle the case where the config is not valid
                throw (new Exception("Open Failed"));
            }
        }

        /// <summary>
        /// Closes the serial port (if open)
        /// </summary>
        /// <param name="close">The Close request - ignored</param>
        [ServiceHandler]
        public void CloseHandler(Close close)
        {
            // Close the serial port
            // If there is an exception, the code will not return and a
            // Fault will be sent back
            ClosePort();
            // Notify subscribers
            SendNotification(_submgrPort, close);
            // Respond to the request
            close.ResponsePort.Post(DefaultUpdateResponseType.Instance);
        }

        /// <summary>
        /// Handles Subscribe messages
        /// </summary>
        /// <param name="subscribe">The Subscribe request</param>
        [ServiceHandler]
        public void SubscribeHandler(Subscribe subscribe)
        {
            // Use the Subscribe Helper
            // Should really check the result to make sure it is successful
            SubscribeHelper(_submgrPort, subscribe.Body, subscribe.ResponsePort);
        }

        /// <summary>
        /// Receives a single byte - Used internally to send notifications
        /// </summary>
        /// <param name="recv">The Receive Byte request</param>
        /// <remarks>This should NEVER be called by applications</remarks>
        [ServiceHandler]
        public void ReceiveByteHandler(ReceiveByte recv)
        {
            _state.LastByteReceived = recv.Body.Data;
            // Send a notification, but only if in asynch mode
            if (_state.Asynchronous)
                SendNotification(_submgrPort, recv);
            recv.ResponsePort.Post(DefaultUpdateResponseType.Instance);
        }


        /// <summary>
        /// Clear Buffers - Clears input, output or both buffers
        /// </summary>
        /// <param name="clear">The Clear request - specifies which buffer(s) to clear using SerialPortClearOptions enum</param>
        [ServiceHandler]
        public void ClearBuffersHandler(ClearBuffers clear)
        {
            if (!_state.IsOpen)
            {
                throw (new Exception("Port not open"));
            }

            switch (clear.Body.buffer)
            {
                case SerialPortClearOptions.Both:
                    sp.DiscardInBuffer();
                    sp.DiscardOutBuffer();
                    break;
                case SerialPortClearOptions.Input:
                    sp.DiscardInBuffer();
                    break;
                case SerialPortClearOptions.Output:
                    sp.DiscardOutBuffer();
                    break;
                default:
                    throw (new Exception("Invalid parameter"));
            }

            // Send a response
            clear.ResponsePort.Post(DefaultSubmitResponseType.Instance);
        }

        /// <summary>
        /// Reads a single byte synchronously
        /// </summary>
        /// <param name="read">The Read Byte request - ignored</param>
        [ServiceHandler]
        public void ReadByteHandler(ReadByte read)
        {
            if (!_state.IsOpen)
            {
                throw (new Exception("Port not open"));
            }

            // NOTE: This will hang forever if the comms link is broken
            // and you have not set a timeout. On the other hand, if there
            // is a timeout then an exception will be raised which is
            // handled automatically by DSS and returned as a Fault.
            int data;
            data = sp.ReadByte();

            // Send the data back
            ReadByteResponseType response = new ReadByteResponseType();
            response.Data = (byte)data;
            read.ResponsePort.Post(response);
        }

        /// <summary>
        /// Writes a single byte synchronously
        /// </summary>
        /// <param name="write">The Write Byte request - contains the data byte to send</param>
        [ServiceHandler]
        public void WriteByteHandler(WriteByte write)
        {
            if (!_state.IsOpen)
            {
                throw (new Exception("Port not open"));
            }

            // NOTE: This might hang forever if the comms link is broken
            // and you have not set a timeout. On the other hand, if there
            // is a timeout then an exception will be raised which is
            // handled automatically by DSS and returned as a Fault.
            byte[] data = new byte[1];
            data[0] = write.Body.Data;
            sp.Write(data, 0, 1);

            // Send back an acknowledgement now that the data is sent
            write.ResponsePort.Post(DefaultSubmitResponseType.Instance);
        }

        /// <summary>
        /// Reads a specified number of bytes synchronously
        /// </summary>
        /// <param name="read">The Read Bytes request - specifies how many bytes to read</param>
        [ServiceHandler]
        public void ReadByteArrayHandler(ReadByteArray read)
        {
            if (!_state.IsOpen)
            {
                throw (new Exception("Port not open"));
            }

            if (read.Body.Count < 1)
                throw(new Exception("Invalid Count"));

            // NOTE: This will hang forever if the comms link is broken
            // and you have not set a timeout. On the other hand, if there
            // is a timeout then an exception will be raised which is
            // handled automatically by DSS and returned as a Fault.
            byte[] data = new byte[read.Body.Count];
            int result;
            try
            {
                result = sp.Read(data, 0, read.Body.Count);
            }
            catch (TimeoutException te)
            {
                // Handle Timeout exceptions separately
                Fault f;
                f = Fault.FromException(new Exception("Timeout"));
                read.ResponsePort.Post(f);
                LogInfo(te);
                return;
            }
            catch (Exception ex)
            {
                // General exception handler
                Fault f;
                f = Fault.FromException(ex);
                read.ResponsePort.Post(f);
                LogError(ex);
                return;
            }

            // Send the data back
            ReadByteArrayResponseType response = new ReadByteArrayResponseType();
            response.DataArray = data;
            read.ResponsePort.Post(response);
        }

        /// <summary>
        /// Writes a specified number of bytes synchronously
        /// </summary>
        /// <param name="write">The Write Bytes request - contains data and count of bytes to write</param>
        [ServiceHandler]
        public virtual IEnumerator<ITask> WriteByteArrayHandler(WriteByteArray write)
        {
            if (!_state.IsOpen)
            {
                throw (new Exception("Port not open"));
            }

            // Check the parameters
            if (write.Body.DataArray == null ||
                write.Body.DataArray.Length < 1 ||
                write.Body.Count < 1 ||
                write.Body.Count > write.Body.DataArray.Length)
                throw(new Exception("Invalid Parameters"));

            // NOTE: This might hang forever if the comms link is broken
            // and you have not set a timeout. On the other hand, if there
            // is a timeout then an exception will be raised which is
            // handled automatically by DSS and returned as a Fault.
            if (_state.Config.InterCharacterDelay > 0)
            {
                for (int i = 0; i < write.Body.Count; i++ )
                {
                    sp.Write(write.Body.DataArray, i, 1);
                    yield return Timeout(_state.Config.InterCharacterDelay);
                }
            }
            else
                sp.Write(write.Body.DataArray,0,write.Body.Count);

            // Send back an acknowledgement now that the data is sent
            write.ResponsePort.Post(DefaultSubmitResponseType.Instance);
            yield break;
        }

        /// <summary>
        /// Reads a "line" of data into a string (terminated by newline)
        /// </summary>
        /// <param name="read">The Read String request - ignored</param>
        [ServiceHandler]
        public void ReadStringHandler(ReadString read)
        {
            if (!_state.IsOpen)
            {
                throw (new Exception("Port not open"));
            }

            // NOTE: This will hang forever if the comms link is broken
            // and you have not set a timeout. On the other hand, if there
            // is a timeout then an exception will be raised which is
            // handled automatically by DSS and returned as a Fault.
            string result;
            try
            {
                result = sp.ReadLine();
            }
            catch (TimeoutException te)
            {
                Fault f;
                f = Fault.FromException(new Exception("Timeout"));
                read.ResponsePort.Post(f);
                LogInfo(te);
                return;
            }
            catch (Exception ex)
            {
                Fault f;
                f = Fault.FromException(ex);
                read.ResponsePort.Post(f);
                LogError(ex);
                return;
            }

            // Send the data back
            ReadStringResponseType response = new ReadStringResponseType();
            response.DataString = result;
            read.ResponsePort.Post(response);
        }

        /// <summary>
        /// Writes a "line" (a string) and appends a newline character
        /// </summary>
        /// <param name="write">The WriteString request - contains the string to send</param>
        [ServiceHandler]
        public virtual IEnumerator<ITask> WriteStringHandler(WriteString write)
        {
            if (!_state.IsOpen)
            {
                throw (new Exception("Port not open"));
            }

            // Check the parameters - An empty string is valid, but not null
            if (write.Body.DataString == null)
                throw (new Exception("Invalid Parameters"));

            // NOTE: This might hang forever if the comms link is broken
            // and you have not set a timeout. On the other hand, if there
            // is a timeout then an exception will be raised which is
            // handled automatically by DSS and returned as a Fault.
            if (_state.Config.InterCharacterDelay > 0)
            {
                byte[] data = new byte[1];
                for (int i = 0; i < write.Body.DataString.Length; i++)
                {
                    data[0] = (byte)write.Body.DataString[i];
                    sp.Write(data, 0, 1);
                    yield return Timeout(_state.Config.InterCharacterDelay);
                }
                sp.WriteLine("");
            }
            else
                sp.WriteLine(write.Body.DataString);

            // Send back an acknowledgement now that the data is sent
            write.ResponsePort.Post(DefaultSubmitResponseType.Instance);
            yield break;
        }

        #endregion

    }
}
