//-----------------------------------------------------------------------
// Copyright  Microsoft Corporation.  All Rights Reserved.
// This code released under the terms of the 
// Microsoft Public License (MS-PL, http://opensource.org/licenses/ms-pl.html)
//
//  $File: SerialPortTypes.cs $ $Revision: 2 $
//-----------------------------------------------------------------------
//

using System;
using System.Collections.Generic;
using System.ComponentModel;
using Microsoft.Ccr.Core;
using Microsoft.Dss.Core.Attributes;
using Microsoft.Dss.ServiceModel.Dssp;
using Microsoft.Dss.ServiceModel.DsspServiceBase;
using W3C.Soap;

// Added references
using System.IO.Ports;


namespace ProMRDS.Robotics.Comms
{
	/// <summary>
	/// SerialPort contract class
	/// </summary>
	public sealed class Contract
	{
		/// <summary>
		/// DSS contract identifer for SerialPort
		/// </summary>
		[DataMember]
		public const string Identifier = "http://www.promrds.com/contracts/2009/12/serialport.html";
	}
	
	/// <summary>
	/// SerialPort state
	/// </summary>
	[DataContract]
	public class SerialPortState
	{
        private bool _openOnStart;
        private bool _asynchronous;
        private SerialPortConfig _config;
        private byte _lastByteReceived;
        private bool _isOpen;

        /// <summary>
        /// Open the COM port when the service starts - must be set in config file
        /// </summary>
        [DataMember]
        public bool OpenOnStart
        {
            get { return _openOnStart; }
            set { _openOnStart = value; }
        }

        /// <summary>
        /// Operate in Asynchronous mode (send notifications for all data received)
        /// </summary>
        [DataMember]
        public bool Asynchronous
        {
            get { return _asynchronous; }
            set { _asynchronous = value; }
        }

        /// <summary>
        /// Configuration parameters for the serial port
        /// </summary>
        [DataMember]
        public SerialPortConfig Config
        {
            get { return _config; }
            set { _config = value; }
        }

        /// <summary>
        /// Last byte received from the serial port
        /// </summary>
        [DataMember]
        public byte LastByteReceived
        {
            get { return _lastByteReceived; }
            set { _lastByteReceived = value; }
        }

        /// <summary>
        /// Indicates if the port is currently open
        /// </summary>
        [DataMember]
        public bool IsOpen
        {
            get { return _isOpen; }
            set { _isOpen = value; }
        }
    }

    /// <summary>
    /// Serial Port Configuration
    /// </summary>
    /// <remarks>
    /// The current config is kept in the State.
    /// Config must be set prior to opening the Serial Port,
    /// which might be from a config file.
    /// </remarks>
    [DataContract]
    public class SerialPortConfig
    {
        // These values are the parameters for SerialPort.Open(),
        // except for PortNumber which has to be converted to a string
        // in the format "COMnn" where nn is the number.
        private int _portNumber;
        private int _baudRate;
        private SerialPortParity _parity;
        private int _dataBits;
        private SerialPortStopBits _stopBits;

        private int _readTimeout;
        private int _writeTimeout;

        private int _interCharacterDelay;

        /// <summary>
        /// COM port number - Defaults to 1
        /// </summary>
        [DataMember]
        public int PortNumber
        {
            get { return _portNumber; }
            set { _portNumber = value; }
        }

        /// <summary>
        /// Baud rate - Defaults to 57600
        /// </summary>
        [DataMember]
        public int BaudRate
        {
            get { return _baudRate; }
            set { _baudRate = value; }
        }

        /// <summary>
        /// Parity - Defaults to None (Use SerialPortParity enum)
        /// </summary>
        [DataMember]
        public SerialPortParity Parity
        {
            get { return _parity; }
            set { _parity = value; }
        }

        /// <summary>
        /// Data bits (Must be 7 or 8) - Defaults to 8
        /// </summary>
        [DataMember]
        public int DataBits
        {
            get { return _dataBits; }
            set { _dataBits = value; }
        }

        /// <summary>
        /// Stop bits - Defaults to One (Use SerialPortStopBits enum)
        /// </summary>
        [DataMember]
        public SerialPortStopBits StopBits
        {
            get { return _stopBits; }
            set { _stopBits = value; }
        }

        /// <summary>
        /// Read Timeout (milliseconds) - Defaults to 5000
        /// </summary>
        [DataMember]
        public int ReadTimeout
        {
            get { return _readTimeout; }
            set { _readTimeout = value; }
        }

        /// <summary>
        /// Write Timeout (milliseconds) - Defaults to 5000
        /// </summary>
        [DataMember]
        public int WriteTimeout
        {
            get { return _writeTimeout; }
            set { _writeTimeout = value; }
        }

        /// <summary>
        /// Delay between characters when sending (milliseconds) - Defaults to 0
        /// </summary>
        /// <remarks>This is not a property of the .NET SerialPort. It is added by this service.</remarks>
        [DataMember]
        public int InterCharacterDelay
        {
            get { return _interCharacterDelay; }
            set { _interCharacterDelay = value; }
        }

    }

    // These enums are redundant because they are already defined
    // for the SerialPort class in .NET. However, they need to be
    // visible in VPL as well.

    /// <summary>
    /// Serial Port Parity enum - Synonyms for .NET Parity enum
    /// </summary>
    [DataContract]
    public enum SerialPortParity
    {
        /// <summary>
        /// Even parity
        /// </summary>
        Even    = Parity.Even,
        /// <summary>
        /// Mark (high) parity
        /// </summary>
        Mark    = Parity.Mark,
        /// <summary>
        /// No parity
        /// </summary>
        None    = Parity.None,
        /// <summary>
        /// Odd parity
        /// </summary>
        Odd     = Parity.Odd,
        /// <summary>
        /// Space (low) parity
        /// </summary>
        Space   = Parity.Space
    }

    /// <summary>
    /// Stop Bits enum - Synonyms for .NET StopBits enum
    /// </summary>
    [DataContract]
    public enum SerialPortStopBits
    {
        /// <summary>
        /// No stop bits
        /// </summary>
        None         = StopBits.None,
        /// <summary>
        /// One stop bit
        /// </summary>
        One          = StopBits.One,
        /// <summary>
        /// One and a half stop bits
        /// </summary>
        OnePointFive = StopBits.OnePointFive,
        /// <summary>
        /// Two stop bits
        /// </summary>
        Two          = StopBits.Two
    }

    /// <summary>
    /// Clear Options enum - Specifies which buffer(s) to clear
    /// </summary>
    [DataContract]
    public enum SerialPortClearOptions
    {
        /// <summary>
        /// Clear both Input and Output Buffers
        /// </summary>
        Both,
        /// <summary>
        /// Clear only Input Buffer
        /// </summary>
        Input,
        /// <summary>
        /// Clear only Output Buffer
        /// </summary>
        Output
    }

    /// <summary>
	/// SerialPort main Operations port
	/// </summary>
	[ServicePort]
	public class SerialPortOperations : PortSet<DsspDefaultLookup, DsspDefaultDrop,
                    Get, Subscribe,
                    SetConfig, Open, Close, ClearBuffers,
                    ReceiveByte,
                    ReadByte, ReadByteArray, ReadString,
                    WriteByte, WriteByteArray, WriteString>
	{
	}

	/// <summary>
	/// SerialPort Get operation -- Returns the State
	/// </summary>
	public class Get : Get<GetRequestType, PortSet<SerialPortState, Fault>>
	{
		/// <summary>
		/// Creates a new instance of Get
		/// </summary>
		public Get()
		{
		}
		
		/// <summary>
		/// Creates a new instance of Get
		/// </summary>
		/// <param name="body">the request message body</param>
		public Get(GetRequestType body)
			: base(body)
		{
		}
		
		/// <summary>
		/// Creates a new instance of Get
		/// </summary>
		/// <param name="body">the request message body</param>
		/// <param name="responsePort">the response port for the request</param>
		public Get(GetRequestType body, PortSet<SerialPortState, Fault> responsePort)
			: base(body, responsePort)
		{
		}
	}

    /// <summary>
    /// SerialPort Subscribe operation -- Register for notifications
    /// </summary>
    public class Subscribe : Subscribe<SubscribeRequestType, PortSet<SubscribeResponseType, Fault>>
    {
        /// <summary>
        /// Creates a new instance of Subscribe
        /// </summary>
        public Subscribe()
        {
        }

        /// <summary>
        /// Creates a new instance of Subscribe
        /// </summary>
        /// <param name="body">the request message body</param>
        public Subscribe(SubscribeRequestType body)
            : base(body)
        {
        }

        /// <summary>
        /// Creates a new instance of Subscribe
        /// </summary>
        /// <param name="body">the request message body</param>
        /// <param name="responsePort">the response port for the request</param>
        public Subscribe(SubscribeRequestType body, PortSet<SubscribeResponseType, Fault> responsePort)
            : base(body, responsePort)
        {
        }
    }

    /// <summary>
    /// SerialPort SetConfig operation -- Configure the port
    /// </summary>
    public class SetConfig : Update<SetConfigRequestType, PortSet<DefaultUpdateResponseType, Fault>>
    {
        /// <summary>
        /// Creates a new instance of SetConfig
        /// </summary>
        public SetConfig()
        {
        }

        /// <summary>
        /// Creates a new instance of SetConfig
        /// </summary>
        /// <param name="body">the request message body</param>
        public SetConfig(SetConfigRequestType body)
            : base(body)
        {
        }
    }

    /// <summary>
    /// Request message for setting config
    /// </summary>
    [DataContract]
    [DataMemberConstructor]
    public class SetConfigRequestType
    {
        // These values are the parameters for SerialPort.Open()
        private SerialPortConfig _config;

        /// <summary>
        /// Config to be set
        /// </summary>
        [DataMember]
        public SerialPortConfig Config
        {
            get { return _config; }
            set { _config = value; }
        }
    }

    /// <summary>
    /// SerialPort Open operation
    /// </summary>
    public class Open : Update<OpenRequestType, PortSet<DefaultUpdateResponseType, Fault>>
    {
        /// <summary>
        /// Creates a new instance of Open
        /// </summary>
        public Open()
        {
        }

        /// <summary>
        /// Creates a new instance of Open
        /// </summary>
        /// <param name="body">the request message body</param>
        public Open(OpenRequestType body)
            : base(body)
        {
        }
    }

    /// <summary>
    /// Open request
    /// </summary>
    [DataContract]
    [DataMemberConstructor]
    public class OpenRequestType
    {
        /// <summary>
        /// Indicates if reads should be asynchronous and use notifications
        /// </summary>
        [DataMember]
        public bool Asynchronous;
    }

    /// <summary>
    /// SerialPort Close operation
    /// </summary>
    public class Close : Update<CloseRequestType, PortSet<DefaultUpdateResponseType, Fault>>
    {
        /// <summary>
        /// Creates a new instance of Close
        /// </summary>
        public Close()
        {
        }

        /// <summary>
        /// Creates a new instance of Close
        /// </summary>
        /// <param name="body">the request message body</param>
        public Close(CloseRequestType body)
            : base(body)
        {
        }
    }

    /// <summary>
    /// Close request -- No fields
    /// </summary>
    [DataContract]
    public class CloseRequestType
    {
    }

    /// <summary>
    /// SerialPort ClearBuffers operation
    /// </summary>
    public class ClearBuffers : Submit<ClearBuffersRequestType, PortSet<DefaultSubmitResponseType, Fault>>
    {
        /// <summary>
        /// Creates a new instance of ClearBuffers
        /// </summary>
        public ClearBuffers()
        {
        }

        /// <summary>
        /// Creates a new instance of ClearBuffers
        /// </summary>
        /// <param name="body">the request message body</param>
        public ClearBuffers(ClearBuffersRequestType body)
            : base(body)
        {
        }
    }

    /// <summary>
    /// ClearBuffers request
    /// </summary>
    [DataContract]
    [DataMemberConstructor]
    public class ClearBuffersRequestType
    {
        /// <summary>
        /// Buffers to be cleared (See SerialPortClearOptions enum)
        /// </summary>
        [DataMember]
        public SerialPortClearOptions buffer;
    }

    /// <summary>
    /// SerialPort ReceiveByte operation -- INTERNAL USE ONLY
    /// </summary>
    /// <remarks>
    /// This operation is used to transfer a byte from the asynchronous
    /// .NET environment into the CCR environment by processing each new
    /// byte via an internal service operation.
    /// </remarks>
    public class ReceiveByte : Update<ReceiveByteRequestType, PortSet<DefaultUpdateResponseType, Fault>>
    {
        /// <summary>
        /// Creates a new instance of ReceiveByte
        /// </summary>
        public ReceiveByte()
        {
        }

        /// <summary>
        /// Creates a new instance of ReceiveByte
        /// </summary>
        /// <param name="body">the request message body</param>
        public ReceiveByte(ReceiveByteRequestType body)
            : base(body)
        {
        }
    }

    /// <summary>
    /// ReceiveByte request -- Contains the byte to process
    /// </summary>
    [DataContract]
    [DataMemberConstructor]
    public class ReceiveByteRequestType
    {
        /// <summary>
        /// Single byte of data received
        /// </summary>
        [DataMember]
        public byte Data;
    }

    /// <summary>
    /// SerialPort ReadByte operation -- Reads one byte synchronously
    /// </summary>
    public class ReadByte : Query<ReadByteRequestType, PortSet<ReadByteResponseType, Fault>>
    {
        /// <summary>
        /// Creates a new instance of ReadByte
        /// </summary>
        public ReadByte()
        {
        }

        /// <summary>
        /// Creates a new instance of ReadByte
        /// </summary>
        /// <param name="body">the request message body</param>
        public ReadByte(ReadByteRequestType body)
            : base(body)
        {
        }
    }

    /// <summary>
    /// ReadByte request -- No fields
    /// </summary>
    [DataContract]
    public class ReadByteRequestType
    {
    }

    /// <summary>
    /// ReadByte response
    /// </summary>
    [DataContract]
    public class ReadByteResponseType
    {
        /// <summary>
        /// Single byte read from the serial port
        /// </summary>
        [DataMember]
        public byte Data;
    }

    /// <summary>
    /// SerialPort ReadByteArray operation -- Reads multiple bytes
    /// </summary>
    public class ReadByteArray : Query<ReadByteArrayRequestType, PortSet<ReadByteArrayResponseType, Fault>>
    {
        /// <summary>
        /// Creates a new instance of ReadByteArray
        /// </summary>
        public ReadByteArray()
        {
        }

        /// <summary>
        /// Creates a new instance of ReadByteArray
        /// </summary>
        /// <param name="body">the request message body</param>
        public ReadByteArray(ReadByteArrayRequestType body)
            : base(body)
        {
        }
    }

    /// <summary>
    /// ReadByte request
    /// </summary>
    [DataContract]
    [DataMemberConstructor]
    public class ReadByteArrayRequestType
    {
        /// <summary>
        /// Number of bytes to read
        /// </summary>
        [DataMember]
        public int Count;
    }

    /// <summary>
    /// ReadByteArray response
    /// </summary>
    [DataContract]
    public class ReadByteArrayResponseType
    {
        /// <summary>
        /// Array of bytes read from serial port
        /// </summary>
        [DataMember]
        public byte[] DataArray;
    }

    /// <summary>
    /// SerialPort ReadString operation -- Does a ReadLine synchronously
    /// </summary>
    /// <remarks>
    /// See the definition of the .NET Serial Port ReadLine method.
    /// Reads up to a NewLine character.
    /// </remarks>
    public class ReadString : Query<ReadStringRequestType, PortSet<ReadStringResponseType, Fault>>
    {
        /// <summary>
        /// Creates a new instance of ReadString
        /// </summary>
        public ReadString()
        {
        }

        /// <summary>
        /// Creates a new instance of ReadString
        /// </summary>
        /// <param name="body">the request message body</param>
        public ReadString(ReadStringRequestType body)
            : base(body)
        {
        }
    }

    /// <summary>
    /// ReadString request -- No fields
    /// </summary>
    [DataContract]
    public class ReadStringRequestType
    {
    }

    /// <summary>
    /// ReadString response
    /// </summary>
    [DataContract]
    public class ReadStringResponseType
    {
        /// <summary>
        /// String read from the serial port
        /// </summary>
        [DataMember]
        public string DataString;
    }

    /// <summary>
    /// SerialPort WriteByte operation - Sends a byte synchronously
    /// </summary>
    public class WriteByte : Submit<WriteByteRequestType, PortSet<DefaultSubmitResponseType, Fault>>
    {
        /// <summary>
        /// Creates a new instance of WriteByte
        /// </summary>
        public WriteByte()
        {
        }

        /// <summary>
        /// Creates a new instance of WriteByte
        /// </summary>
        /// <param name="body">the request message body</param>
        public WriteByte(WriteByteRequestType body)
            : base(body)
        {
        }
    }

    /// <summary>
    /// WriteByte request
    /// </summary>
    [DataContract]
    [DataMemberConstructor]
    public class WriteByteRequestType
    {
        /// <summary>
        /// Single byte to write to the serial port
        /// </summary>
        [DataMember]
        public byte Data;
    }

    /// <summary>
    /// SerialPort WriteByteArray operation - Sends multiple bytes
    /// </summary>
    public class WriteByteArray : Submit<WriteByteArrayRequestType, PortSet<DefaultSubmitResponseType, Fault>>
    {
        /// <summary>
        /// Creates a new instance of WriteByteArray
        /// </summary>
        public WriteByteArray()
        {
        }

        /// <summary>
        /// Creates a new instance of WriteByteArray
        /// </summary>
        /// <param name="body">the request message body</param>
        public WriteByteArray(WriteByteArrayRequestType body)
            : base(body)
        {
        }
    }

    /// <summary>
    /// WriteByteArray request
    /// </summary>
    [DataContract]
    [DataMemberConstructor]
    public class WriteByteArrayRequestType
    {
        /// <summary>
        /// Number of bytes to write - Does not have to be the full array
        /// </summary>
        [DataMember]
        public int Count;

        /// <summary>
        /// Byte array to write to the serial port
        /// </summary>
        [DataMember]
        public byte[] DataArray;
    }

    /// <summary>
    /// SerialPort WriteString operation - Sends a string
    /// </summary>
    /// <remarks>
    /// See the definition of the .NET Serial Port WriteLine method.
    /// Sends a NewLine character after the string.
    /// The config has an Intercharacter Delay to slow down transmission
    /// for slow devices.
    /// </remarks>
    public class WriteString : Submit<WriteStringRequestType, PortSet<DefaultSubmitResponseType, Fault>>
    {
        /// <summary>
        /// Creates a new instance of WriteString
        /// </summary>
        public WriteString()
        {
        }

        /// <summary>
        /// Creates a new instance of WriteString
        /// </summary>
        /// <param name="body">the request message body</param>
        public WriteString(WriteStringRequestType body)
            : base(body)
        {
        }
    }

    /// <summary>
    /// WriteString request
    /// </summary>
    [DataContract]
    [DataMemberConstructor]
    public class WriteStringRequestType
    {
        /// <summary>
        /// String to be written to the serial port
        /// </summary>
        [DataMember]
        public string DataString;
    }

    /// <summary>
    /// Response to a WriteString request -- No fields
    /// </summary>
    [DataContract]
    public class WriteStringResponseType
    {
    }

}

